/* X Letters - Catch falling words */
/* Copyright (C) 1998 by David A. Madore (david.madore@ens.fr) */

/*
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version
 * 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See
 * the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/* General include files */
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <time.h>
#include <pwd.h>
#include <sys/types.h>
#include <unistd.h>

/* Xlib include files */
#include <X11/Xlib.h>

/* Xt include files */
#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>

/* Xaw include files */
#include <X11/Xaw/Box.h>
#include <X11/Xaw/Label.h>
#include <X11/Xaw/Command.h>
#include <X11/Xaw/Toggle.h>

/* Configuration file */
#include "config.h"

/* Game variables */
int level;
long score;
int lives;
char bonuslevel;

typedef struct word_state_s {
  char *s; /* The actual word - malloc()ated */
  int len; /* The length of that word */
  int typed; /* Number of characters correctly typed */
  int width,twidth; /* Width in pixels of whole string and typed part */
  long x; /* x position of left-most point, in pixels */
  long y; /* y position of baseline, in *millipixels* */
  long rate; /* Descent rate in millipixels per time unit */
} word_state_t;

/* The words that are falling */
word_state_t words[MAXWORDS];
/* The number of words that are falling */
int nb_words;

#define MAX_WORDSIZE 1023

/* Whether the game is paused */
char paused;

/* Times are measured in fundamental time units (normally 75ms) */
unsigned long current_time;
unsigned long lastword_time;
unsigned long level_time;

/* The word file */
FILE *wordf;
/* The length of the word file in characters */
long wordf_len;

/* X related variables */
char *fallback_resources[] = {
  "*input: True",
  "*background: Grey51",
  "*foreground: White",
  "*Box.background: DarkSlateGrey",
  "*Label.foreground: White",
  "*label.foreground: Green",
  "*Command.foreground: Yellow",
  "*Toggle.foreground: Yellow",
  "*quitButton.label: Quit",
  "*pauseButton.label: Pause",
  "*quitButton.accelerators:"
  "<KeyPress>Escape: set() notify()",
  "*pauseButton.accelerators:"
  "<KeyPress>Pause: toggle() notify()\\n\\\n"
  "<KeyPress>Tab: toggle() notify()",
  NULL
};

typedef struct app_res_val_s {
  XFontStruct *word_font;
  Pixel word_color,typed_color;
} app_res_val_t;

app_res_val_t app_res_val;

XtResource app_res_tab[] = {
  {"wordFont", XtCFont, XtRFontStruct, sizeof(XFontStruct *),
   XtOffsetOf(app_res_val_t,word_font),
   XtRString, FONTNAME
  },
  {"wordColor", XtCForeground, XtRPixel, sizeof(Pixel),
   XtOffsetOf(app_res_val_t,word_color),
   XtRString, WORD_COLOR
  },
  {"typedColor", "HighlightColor", XtRPixel, sizeof(Pixel),
   XtOffsetOf(app_res_val_t,typed_color),
   XtRString, TYPED_COLOR
  },
};

XrmOptionDescRec options[] = {
  {"-wfn", "*wordFont", XrmoptionSepArg, NULL},
  {"-wc", "*wordColor", XrmoptionSepArg, NULL},
  {"-tc", "*typedColor", XrmoptionSepArg, NULL},
  {"-gbg", "*gameSpace.background", XrmoptionSepArg, NULL},
};

XtIntervalId timeout;

Widget toplevel,ground_box,label;
Widget quit_button,pause_button;
Widget lives_label,score_label,level_label;
Widget game_space;

XtAppContext app_context;

Display *dpy;
Window win;
GC gc0,gc1,gc2;
XFontStruct *fnt_info;
Font fnt;

/* High score table */
typedef struct hstable_entry_s {
  char name[1024];
  uid_t uid;
  int level;
  long score;
  unsigned long dur;
} hstable_entry_t;

hstable_entry_t hstable[SCORELEN];

uid_t euid;
gid_t egid;

void
print_hstable_entry(hstable_entry_t *ent,int n)
{
  int i,f;
  if (n==-1) printf("You: ");
  else printf("%4d ",n+1);
  f=1;
  for (i=0;i<20;i++) {
    if (f&&(ent->name[i]==0)) f=0;
    if (f) {
      if (ent->name[i]=='_') printf(" ");
      else printf("%c",ent->name[i]);
    } else printf(" ");
  }
  printf("%4d%6ld%3lu'%02lu\n",ent->level,ent->score,
	 ent->dur/60000,(ent->dur/1000)%60);
}

void
show_hstable(void)
{
  hstable_entry_t yours;
  FILE *hsf;
  int i,replace,place;
  struct passwd *pwent;
  yours.uid = getuid();
  pwent=getpwuid(yours.uid);
  strncpy(yours.name,pwent->pw_gecos,1000);
  yours.name[1000]=0;
  for (i=0;i<1000;i++) {
    if (yours.name[i]==' ') yours.name[i]='_';
    if (yours.name[i]==',') yours.name[i]=0;
  }
  yours.level=level;
  yours.score=score;
  yours.dur=current_time*BASETIME;
  printf("     Name                Lev.Score Time\n");
  print_hstable_entry(&yours,-1);
  seteuid(euid);
  setegid(egid);
  hsf = fopen(SCOREFILE,"r+");
  if (hsf==NULL) {
    hsf = fopen(SCOREFILE,"w+");
    if (hsf==NULL) {
      perror("Can't open high score file");
      exit(1);
    }
  }
  rewind(hsf);
  for (i=0;i<SCORELEN;i++) {
    unsigned long x;
    if (fscanf(hsf,"%s %lu %d %ld %lu\n",hstable[i].name,
	       &x,&hstable[i].level,&hstable[i].score,
	       &hstable[i].dur)==5) {
      /* Beware: the above lines contain a buffer overrun.  Don't mess
       * up with the high score tables!  In particular, if the program
       * is made suid or sgid, only people with the corresponding uid
       * or gid should be allowed to modify the table. */
      hstable[i].uid=x;
    } else {
      strcpy(hstable[i].name,"Joe_Player");
      hstable[i].uid=9999;
      hstable[i].level=0;
      hstable[i].score=0;
      hstable[i].dur=0;
    }
  }
#if MULTENT
  replace=SCORELEN-1;
#else
  for (replace=0;replace<SCORELEN-1;replace++) {
    if (hstable[replace].uid==yours.uid)
      break;
  }
#endif
  if (yours.score>hstable[replace].score) {
    printf("You made it into the high score file!\n");
    for (place=0;place<SCORELEN;place++) {
      if (yours.score>hstable[place].score)
	break;
    }
    for (i=replace;i>place;i--) {
      hstable[i]=hstable[i-1];
    }
    hstable[place]=yours;
  }
  printf("HIGH SCORES:\n");
  for (i=0;i<SCORELEN;i++)
    print_hstable_entry(hstable+i,i);
  rewind(hsf);
  for (i=0;i<SCORELEN;i++) {
    fprintf(hsf,"%s %lu %d %ld %lu\n",hstable[i].name,
	    (long)hstable[i].uid,hstable[i].level,hstable[i].score,
	    hstable[i].dur);
  }
  fclose(hsf);
  exit(0);
}

long
level_speed(int lev)
     /* Return the base descent rate (in millipixels per time unit)
      * for the level. */
{
  return 900+80*lev;
}

long
level_rate(int lev)
     /* How long to wait before to drop a new word on the given
      * level */
{
  int l;
  l = ((lev<20)?lev:20);
  return 10+(50000-l*2500)/level_speed(lev);
}

void
show_lives(void)
     /* Update the lives label. */
{
  char s[256];
  sprintf(s,"Lives: %d",lives);
  XtVaSetValues(lives_label,
		XtNlabel,s,
		NULL);
}

void
show_score(void)
     /* Update the score label. */
{
  char s[256];
  sprintf(s,"Score: %ld",score);
  XtVaSetValues(score_label,
		XtNlabel,s,
		NULL);
}

void
show_level(void)
     /* Update the level label. */
{
  char s[256];
  sprintf(s,"Level: %d",level);
  XtVaSetValues(level_label,
		XtNlabel,s,
		NULL);
}

void
add_score(long n)
     /* Increase player's score. */
{
  long newscore;
  newscore=score+n;
  if ((newscore/2500)>(score/2500)) {
    lives++;
    show_lives();
  }
  score=newscore;
  show_score();
}

void
lose_life(void)
     /* Remove a life from the player. */
{
  if (lives==1) {
    show_hstable();
  }
  lives--;
  show_lives();
}

void expose_proc(Widget w,XtPointer client_data,XEvent *evt,Boolean *cnt);

void
clear_all(void)
     /* Remove all falling words. */
{
  int i;
  Boolean garbage;
  for (i=0;i<nb_words;i++)
    free(words[i].s);
  nb_words=0;
  expose_proc(game_space,NULL,NULL,&garbage);
}

void
new_level(void)
     /* Switch to next level. */
{
  clear_all();
  if (bonuslevel) {
    level++;
    bonuslevel=0;
    show_level();
  } else bonuslevel=1;
  level_time=current_time;
  lastword_time=-level_rate(level);
}

void
draw_word(word_state_t *w)
     /* Base routine for drawing a word - or erasing it. */
{
  XDrawString(dpy,win,gc1,w->x,w->y/1000,
	      w->s,w->typed);
  XDrawString(dpy,win,gc2,w->x+w->twidth,w->y/1000,
	      w->s+w->typed,w->len-w->typed);
}

void timer_proc(XtPointer client_data,XtIntervalId *id);

void
retime(void)
     /* Make sure the game is running. */
{
  paused = 0;
  timeout = XtAppAddTimeOut(app_context,BASETIME,timer_proc,NULL);
}

void
resume(void)
     /* Resume the game if it was paused. */
{
  if (paused)
    retime();
}

void
pauze(void)
     /* Pause the game if it wasn't. */
{
  if (!paused) {
    XtRemoveTimeOut(timeout);
    paused = 1;
  }
}

void
quit_callback(Widget w,XtPointer client_data,XtPointer call_data)
     /* Callback called by the quit button */
{
  show_hstable();
}

void
pause_callback(Widget w,XtPointer client_data,XtPointer call_data)
     /* Callback called by the pause button */
{
  char b;
  XtVaGetValues(w,
		XtNstate,&b,
		NULL);
  if (b)
    pauze();
  else
    resume();
}

void
x_resume(void)
     /* Resume the game, untoggling the Pause button if necessary. */
{
  if (!paused) return;
  XtVaSetValues(pause_button,
		XtNstate,False,
		NULL);
  resume();
}

void
new_word(void)
     /* Make a new word fall. */
{
  if (nb_words>=MAXWORDS) return;
  if (bonuslevel) {
    int i;
    words[nb_words].len = random()%5+5;
    words[nb_words].s = malloc(words[nb_words].len+1);
    if (words[nb_words].s==NULL) {
      fprintf(stderr,"Out of memory!\n");
      exit(1);
    }
    for (i=0;i<words[nb_words].len;i++)
      words[nb_words].s[i] = 33+random()%94;
    words[nb_words].s[words[nb_words].len] = 0;
  } else {
    char s[MAX_WORDSIZE+1];
    int ch;
    do {
      if (fseek(wordf,random()%wordf_len,SEEK_SET)<0) {
	perror("Can't seek " WORDFILE);
	exit(1);
      }
      ch = fgetc(wordf);
    } while (ch!='\n');
    fgets(s,MAX_WORDSIZE,wordf);
    words[nb_words].len = strlen(s)-1;
    words[nb_words].s = malloc(words[nb_words].len+1);
    if (words[nb_words].s==NULL) {
      fprintf(stderr,"Out of memory!\n");
      exit(1);
    }
    memcpy(words[nb_words].s,s,words[nb_words].len);
    words[nb_words].s[words[nb_words].len] = 0;
  }
  words[nb_words].typed = 0;
  {
    int dir_ret,f_a_ret,f_d_ret;
    XCharStruct overall;
    XTextExtents(fnt_info,words[nb_words].s,words[nb_words].len,
		 &dir_ret,&f_a_ret,&f_d_ret,&overall);
    words[nb_words].width = overall.width;
  }
  words[nb_words].twidth = 0;
  if (words[nb_words].width<WIDTH)
    words[nb_words].x = random()%(WIDTH-words[nb_words].width);
  else words[nb_words].x = 0;
  words[nb_words].y = 0;
  words[nb_words].rate = level_speed(level)*7/(2+words[nb_words].len);
#if NOCOLLIDE
  {
    int i;
    int left;
    int right;
    char redo;
    left=words[nb_words].x;
    right=words[nb_words].x+words[nb_words].width;
    do {
      redo=0;
      for (i=0;i<nb_words;i++) {
	if ((words[i].x<right)&&(words[i].x+words[i].width>=left)) {
	  if (words[i].rate<words[nb_words].rate)
	    words[i].rate=words[nb_words].rate;
	  if (words[i].x<left) {
	    left=words[i].x;
	    redo=1;
	  }
	  if (words[i].x+words[i].width>right) {
	    right=words[i].x+words[i].width;
	    redo=1;
	  }
	}
      }
    } while (redo);
  }
#endif
  draw_word(words+nb_words);
  lastword_time=current_time;
  nb_words++;
}

void
calc_twidth(word_state_t *w)
     /* Calculate the twidth field of the word state structure. */
{
  int dir_ret,f_a_ret,f_d_ret;
  XCharStruct overall;
  XTextExtents(fnt_info,w->s,w->typed,
	       &dir_ret,&f_a_ret,&f_d_ret,&overall);
  w->twidth = overall.width;
}

char
key_typed(word_state_t *w)
     /* Mark a new correct key as typed in the given word. */
{
  w->typed++;
  calc_twidth(w);
  return (w->typed>=w->len);
}

void
key_untyped(word_state_t *w,char ch)
     /* Determine what happens when a wrong key has been pressed. */
{
#if CONTEXT
  int t0,i;
  t0=w->typed;
  for (;w->typed>=0;w->typed--) {
    if (w->s[w->typed]!=ch)
      continue;
    for (i=0;i<w->typed;i++)
      if (w->s[w->typed-1-i]!=w->s[t0-1-i])
	break;
    if (i>=w->typed)
      break;
    /* Sometimes I think I should participate in IOCCC's... */
  }
  w->typed++;
  calc_twidth(w);
#else
  w->typed = 0;
  w->twidth = 0;
#endif
}

void
timer_proc(XtPointer client_data,XtIntervalId *id)
     /* Timeout handler called every time unit (when the game is
      * unpaused) */
{
  int i;
  current_time++;
  if (current_time>=level_time+LEVELDUR)
    new_level();
  for (i=0;i<nb_words;i++) {
    word_state_t w;
    w = words[i];
    words[i].y+=words[i].rate;
    if (words[i].y>HEIGHT*1000) {
      if (!bonuslevel) {
	int j;
	nb_words--;
	for (j=i;j<nb_words;j++)
	  words[j]=words[j+1];
	i--; /* YUCK ALERT */
	draw_word(&w);
	free(w.s);
	lose_life();
      } else {
	new_level();
	break;
      }
    } else {
      draw_word(words+i);
      draw_word(&w);
    }
  }
  if (current_time>=lastword_time+level_rate(level)) {
    new_word();
  }
  retime();
}

#define KBD_1BUF_SIZE 256 /* Rather ridiculous: 1 should be enough */

void
key_proc(Widget w,XtPointer client_data,XEvent *evt,Boolean *cnt)
     /* Event handler called when a key is pressed */
{
  int i;
  char buf[KBD_1BUF_SIZE];
  int buflen;
  int p;
  KeySym keysym;
  buflen=XLookupString(&evt->xkey,buf,KBD_1BUF_SIZE,&keysym,NULL);
  for (p=0;p<buflen;p++) {
    if ((buf[p]<=32)||(buf[p]>=127))
      continue;
    x_resume();
    for (i=0;i<nb_words;i++) {
      word_state_t w;
      w = words[i];
      if (words[i].s[words[i].typed]==buf[p]) {
	if (key_typed(words+i)) {
	  int j;
	  nb_words--;
	  for (j=i;j<nb_words;j++)
	    words[j]=words[j+1];
	  i--; /* YUCK ALERT */
	  draw_word(&w);
	  free(w.s);
	  add_score((w.y<HEIGHT*200)?8+w.len:5+w.len);
	} else {
	  draw_word(words+i);
	  draw_word(&w);
	}
      } else {
	key_untyped(words+i,buf[p]);
	draw_word(words+i);
	draw_word(&w);
      }
    }
  }
  *cnt=True;
}

void
expose_proc(Widget w,XtPointer client_data,XEvent *evt,Boolean *cnt)
     /* Event handler called when an Expose event is received */
{
  int i;
  XFillRectangle(dpy,win,gc0,0,0,WIDTH,HEIGHT);
  for (i=0;i<nb_words;i++)
    draw_word(words+i);
  *cnt=False;
}

/* The main function (yeah, this is too long) */

int
main(int argc,char **argv)
{
  /* Save uid and gid for later */
  euid = geteuid();
  egid = getegid();
  seteuid(getuid());
  setegid(getgid());
  /* Randomize. */
  srandom(time(NULL));
  /* Open the word file. */
  wordf = fopen(WORDFILE,"r");
  if (wordf==NULL) {
    perror("Can't open " WORDFILE);
    exit(1);
  }
  if (fseek(wordf,0,SEEK_END)<0) {
    perror("Can't seek " WORDFILE);
    exit(1);
  }
  wordf_len = ftell(wordf);
  /* Read application resources. */
  toplevel = XtVaAppInitialize(&app_context,APP_CLASS,
			       options,XtNumber(options),
			       &argc,argv,fallback_resources,
			       NULL);
  if (argc>1) {
    /* We don't deal properly with GNU long options, but I don't
     * intend to link with GNU getopt simply so people are able to
     * type --vers instead of --version */
    if ((strcmp(argv[1],"-version")==0)||(strcmp(argv[1],"-v")==0)
	||(strcmp(argv[1],"-V")==0)
	||(strcmp(argv[1],"--version")==0)||(strcmp(argv[1],"--ver")==0)) {
      fprintf(stderr,"xletters - catch falling words\n");
      fprintf(stderr,"Version 1.0.0\n");
      fprintf(stderr,"Copyright (C) 1998 by "
	      "David A. Madore (david.madore@ens.fr)\n");
      fprintf(stderr,"xletters comes with ABSOLUTELY NO WARRANTY:\n");
      fprintf(stderr,"see the GNU general public license for more details.\n");
      exit(0);
    } else if ((strcmp(argv[1],"-help")==0)||(strcmp(argv[1],"-h")==0)
	       ||(strcmp(argv[1],"--help")==0)) {
      fprintf(stderr,"Usage is:\n");
      fprintf(stderr,"%s [-version] [-help] [-bg color] [-fg color]"
	      "[-fn font]\n",argv[0]);
      fprintf(stderr,"[-name name] [-title string] [-geometry geom]"
	      "[-display disp]\n");
      fprintf(stderr,"[-xrm resource] [-wfn font] [-wc color]"
	      "[-tc color] [-gbg color]\n");
      exit(0);
    } else {
      fprintf(stderr,"Unrecognized option %s\n",argv[1]);
      fprintf(stderr,"Type %s -help for help\n",argv[0]);
      exit(1);
    }
  }
  XtGetApplicationResources(toplevel,&app_res_val,app_res_tab,
			    XtNumber(app_res_tab),NULL,0);
  /* Create widgets. */
  ground_box = XtVaCreateManagedWidget("groundBox",boxWidgetClass,toplevel,
				       NULL);
  label = XtVaCreateManagedWidget("label",labelWidgetClass,ground_box,
				  XtNlabel,"xletters",
				  NULL);
  quit_button = XtVaCreateManagedWidget("quitButton",commandWidgetClass,
					ground_box,
					NULL);
  pause_button = XtVaCreateManagedWidget("pauseButton",toggleWidgetClass,
					 ground_box,
					 NULL);
  lives_label = XtVaCreateManagedWidget("livesLabel",labelWidgetClass,
					ground_box,
					NULL);
  score_label = XtVaCreateManagedWidget("scoreLabel",labelWidgetClass,
					ground_box,
					NULL);
  level_label = XtVaCreateManagedWidget("levelLabel",labelWidgetClass,
					ground_box,
					NULL);
  game_space = XtVaCreateManagedWidget("gameSpace",coreWidgetClass,ground_box,
				       XtNwidth,WIDTH,
				       XtNheight,HEIGHT,
				       NULL);
  XtRealizeWidget(toplevel);
  /* Xlib miscellanea */
  dpy = XtDisplay(game_space);
  win = XtWindow(game_space);
  gc0 = XCreateGC(dpy,win,0,NULL);
  {
    XGCValues val;
    val.function = GXxor;
    gc1 = XCreateGC(dpy,win,GCFunction,&val);
    gc2 = XCreateGC(dpy,win,GCFunction,&val);
  }
  {
    Pixel bgc,c1,c2;
    XtVaGetValues(game_space,
		  XtNbackground,&bgc,
		  NULL);
    XSetForeground(dpy,gc0,bgc);
    c1 = app_res_val.typed_color;
    c2 = app_res_val.word_color;
    XSetForeground(dpy,gc1,bgc^c1);
    XSetForeground(dpy,gc2,bgc^c2);
  }
  fnt_info = app_res_val.word_font;
  fnt = fnt_info->fid;
  XSetFont(dpy,gc0,fnt);
  XSetFont(dpy,gc1,fnt);
  XSetFont(dpy,gc2,fnt);
  /* Setup callbacks and handlers */
  retime();
  XtAddCallback(quit_button,XtNcallback,quit_callback,NULL);
  XtAddCallback(pause_button,XtNcallback,pause_callback,NULL);
  XtAddEventHandler(game_space,KeyPressMask,False,key_proc,NULL);
  XtAddEventHandler(game_space,ExposureMask,False,expose_proc,NULL);
  XtInstallAllAccelerators(game_space,ground_box);
  XtInstallAllAccelerators(ground_box,ground_box);
  /* Etcetera */
  lives = 5;
  score = 0;
  level = 1;
  bonuslevel = 0;
  lastword_time = -level_rate(level);
  show_lives();
  show_score();
  show_level();
  XtAppMainLoop(app_context);
  return 0;
}
